from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.types import FSharpRef
from ..arctrl_core.Helper.collections_ import StringDictionary_ofSeq
from ..arctrl_core.Helper.hash_codes import (box_hash_seq, merge_hashes, hash_1 as hash_1_1)
from ..arctrl_core.Helper.regex import ActivePatterns__007CRegex_007C__007C
from ..fable_library.map_util import (add_to_dict, get_item_from_dict, try_get_value)
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.reg_exp import (get_item, groups)
from ..fable_library.seq import (reduce, map, sort, try_pick)
from ..fable_library.string_ import (starts_with_exact, to_fail, printf, replace)
from ..fable_library.types import (to_string, Array)
from ..fable_library.uri import Uri
from ..fable_library.util import (IEnumerable_1, get_enumerator, dispose, safe_hash, compare_primitives, string_hash)

IRIHelper_compactIRIRegex: str = "(?<prefix>.*):(?<suffix>[^\\/][^\\/].*)"

def IRIHelper__007CCompactIri_007C__007C(term_defition: str) -> tuple[str, str] | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C(IRIHelper_compactIRIRegex, term_defition)
    if active_pattern_result is not None:
        result: Any = active_pattern_result
        return (get_item(groups(result), "prefix") or "", get_item(groups(result), "suffix") or "")

    else: 
        return None



def IRIHelper_combine(base_iri: str, relative: str) -> str:
    if True if starts_with_exact(relative, "http://") else starts_with_exact(relative, "https://"):
        return relative

    else: 
        return to_string(Uri.create(Uri.create(base_iri), relative))



def IRIHelper_combineOptional(base_iri: str | None=None, relative: str | None=None) -> str | None:
    if base_iri is None:
        if relative is not None:
            r_1: str = relative
            return r_1

        else: 
            return None


    elif relative is None:
        b_1: str = base_iri
        return b_1

    else: 
        b: str = base_iri
        r: str = relative
        return IRIHelper_combine(b, r)



def _expr1022() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDContext", None, LDContext)


class LDContext:
    def __init__(self, mappings: Any | None=None, base_contexts: Array[LDContext] | None=None) -> None:
        self.base_contexts_004041: Array[LDContext] = default_arg(base_contexts, [])
        self.name: str | None = None
        self.mappings_004045: Any = dict([]) if (mappings is None) else mappings
        self.reverse_mappings: Any = dict([])
        self.compact_reverse_mappings: Any = dict([])
        enumerator: Any = get_enumerator(self.mappings_004045)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                kvp: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                self.add_reverse_mapping(kvp[0], kvp[1])

        finally: 
            dispose(enumerator)


    @property
    def Mappings(self, __unit: None=None) -> Any:
        this: LDContext = self
        return this.mappings_004045

    @property
    def BaseContexts(self, __unit: None=None) -> Array[LDContext]:
        this: LDContext = self
        return this.base_contexts_004041

    @BaseContexts.setter
    def BaseContexts(self, value: Array[LDContext]) -> None:
        this: LDContext = self
        this.base_contexts_004041 = value

    @property
    def Name(self, __unit: None=None) -> str | None:
        this: LDContext = self
        return this.name

    @Name.setter
    def Name(self, value: str | None=None) -> None:
        this: LDContext = self
        this.name = value

    def AddMapping(self, term: str, definition: str) -> None:
        this: LDContext = self
        try: 
            key: str = term
            value: str = definition
            dict_1: Any = this.mappings_004045
            if key in dict_1:
                dict_1[key] = value

            else: 
                add_to_dict(dict_1, key, value)

            this.add_reverse_mapping(term, definition)

        except Exception as ex:
            arg_2: str = str(ex)
            to_fail(printf("Failed to add mapping to context: %s -> %s: %s"))(term)(definition)(arg_2)


    def TryResolveTerm(self, term: str) -> str | None:
        this: LDContext = self
        def mapping(term_1: str) -> str | None:
            return this.try_find_term(term_1)

        return reduce(IRIHelper_combineOptional, map(mapping, term.split(":"))) if (term.find(":") >= 0) else this.try_find_term(term)

    def TryGetTerm(self, iri: str) -> str | None:
        this: LDContext = self
        return this.try_find_iri(iri)

    def PropertyNamesMatch(self, p1: str, p2: str) -> bool:
        this: LDContext = self
        if p1 == p2:
            return True

        else: 
            p1def: str | None = this.TryResolveTerm(p1)
            p2def: str | None = this.TryResolveTerm(p2)
            def _arrow1005(__unit: None=None) -> bool:
                p2def_2: str = p2def
                return p1 == p2def_2

            def _arrow1006(__unit: None=None) -> bool:
                p1def_2: str = p1def
                return p1def_2 == p2

            def _arrow1007(__unit: None=None) -> bool:
                p1def_1: str = p1def
                p2def_1: str = p2def
                return p1def_1 == p2def_1

            return (_arrow1005() if (p2def is not None) else False) if (p1def is None) else (_arrow1006() if (p2def is None) else _arrow1007())


    @staticmethod
    def from_mapping_seq(mappings: IEnumerable_1[tuple[str, str]]) -> LDContext:
        return LDContext(StringDictionary_ofSeq(mappings))

    @staticmethod
    def combine_in_place(base_context: LDContext, specific_context: LDContext) -> LDContext:
        (specific_context.BaseContexts.append(base_context))
        return specific_context

    @staticmethod
    def combine(base_context: LDContext, specific_context: LDContext) -> LDContext:
        return LDContext(None, [specific_context, base_context])

    @staticmethod
    def try_combine_optional(base_context: LDContext | None=None, specific_context: LDContext | None=None) -> LDContext | None:
        def _arrow1008(__unit: None=None) -> LDContext | None:
            s_1: LDContext = specific_context
            return s_1

        def _arrow1009(__unit: None=None) -> LDContext | None:
            f_1: LDContext = base_context
            return f_1

        def _arrow1010(__unit: None=None) -> LDContext | None:
            f: LDContext = base_context
            s: LDContext = specific_context
            return LDContext.combine(f, s)

        return (_arrow1008() if (specific_context is not None) else None) if (base_context is None) else (_arrow1009() if (specific_context is None) else _arrow1010())

    def ShallowCopy(self, __unit: None=None) -> LDContext:
        this: LDContext = self
        new_mappings: Any = dict([])
        enumerator: Any = get_enumerator(this.mappings_004045)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                kvp: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                add_to_dict(new_mappings, kvp[0], kvp[1])

        finally: 
            dispose(enumerator)

        return LDContext(new_mappings, this.base_contexts_004041)

    def DeepCopy(self, __unit: None=None) -> LDContext:
        this: LDContext = self
        new_mappings: Any = dict([])
        enumerator: Any = get_enumerator(this.mappings_004045)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                kvp: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                add_to_dict(new_mappings, kvp[0], kvp[1])

        finally: 
            dispose(enumerator)

        new_base_contexts: Array[LDContext] = []
        enumerator_1: Any = get_enumerator(this.base_contexts_004041)
        try: 
            while enumerator_1.System_Collections_IEnumerator_MoveNext():
                ctx: LDContext = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                (new_base_contexts.append(ctx.DeepCopy()))

        finally: 
            dispose(enumerator_1)

        return LDContext(new_mappings, new_base_contexts)

    def StructurallyEquals(self, other: LDContext) -> bool:
        this: LDContext = self
        return safe_hash(this) == safe_hash(other)

    def ReferenceEquals(self, other: LDContext) -> bool:
        this: LDContext = self
        return this is other

    def __eq__(self, other: Any=None) -> bool:
        this: LDContext = self
        return this.StructurallyEquals(other) if isinstance(other, LDContext) else False

    def __hash__(self, __unit: None=None) -> int:
        this: LDContext = self
        def mapping(k: str) -> int:
            return merge_hashes(hash_1_1(k), hash_1_1(get_item_from_dict(this.Mappings, k)))

        class ObjectExpr1011:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        mappings_hash: int = box_hash_seq(map(mapping, sort(this.Mappings.keys(), ObjectExpr1011())))
        name_hash: int
        match_value: str | None = this.Name
        name_hash = 0 if (match_value is None) else string_hash(match_value)
        base_contexts_hash: int = (0 if (len(this.base_contexts_004041) == 0) else reduce(merge_hashes, map(safe_hash, this.base_contexts_004041))) or 0
        return merge_hashes(merge_hashes(mappings_hash, name_hash), base_contexts_hash)

    def System_ICloneable_Clone(self, __unit: None=None) -> Any:
        this: LDContext = self
        return this.DeepCopy()

    def add_reverse_mapping(self, key: str, value: str) -> None:
        this: LDContext = self
        http: str = replace(value, "https://", "http://")
        https: str = replace(value, "http://", "https://")
        key_1: str = http
        value_1: str = key
        dict_1: Any = this.reverse_mappings
        if key_1 in dict_1:
            dict_1[key_1] = value_1

        else: 
            add_to_dict(dict_1, key_1, value_1)

        key_2: str = https
        value_2: str = key
        dict_1_1: Any = this.reverse_mappings
        if key_2 in dict_1_1:
            dict_1_1[key_2] = value_2

        else: 
            add_to_dict(dict_1_1, key_2, value_2)

        active_pattern_result: tuple[str, str] | None = IRIHelper__007CCompactIri_007C__007C(value)
        if active_pattern_result is not None:
            prefix: str = active_pattern_result[0]
            suffix: str = active_pattern_result[1]
            key_3: str = prefix
            value_3: tuple[str, str] = (suffix, key)
            dict_1_2: Any = this.compact_reverse_mappings
            if key_3 in dict_1_2:
                dict_1_2[key_3] = value_3

            else: 
                add_to_dict(dict_1_2, key_3, value_3)

            match_value: str | None
            pattern_input: tuple[bool, str]
            out_arg: str = None
            def _arrow1012(__unit: None=None) -> str:
                return out_arg

            def _arrow1013(v: str) -> None:
                nonlocal out_arg
                out_arg = v

            pattern_input = (try_get_value(this.mappings_004045, prefix, FSharpRef(_arrow1012, _arrow1013)), out_arg)
            match_value = pattern_input[1] if pattern_input[0] else None
            if match_value is None:
                pass

            else: 
                iri: str = IRIHelper_combine(match_value, suffix)
                this.add_reverse_mapping(key, iri)


        else: 
            match_value_1: tuple[str, str] | None
            pattern_input_1: tuple[bool, tuple[str, str]]
            out_arg_1: tuple[str, str] = None
            def _arrow1014(__unit: None=None) -> tuple[str, str]:
                return out_arg_1

            def _arrow1015(v_2: tuple[str, str]) -> None:
                nonlocal out_arg_1
                out_arg_1 = v_2

            pattern_input_1 = (try_get_value(this.compact_reverse_mappings, key, FSharpRef(_arrow1014, _arrow1015)), out_arg_1)
            match_value_1 = pattern_input_1[1] if pattern_input_1[0] else None
            if match_value_1 is None:
                pass

            else: 
                term: str = match_value_1[1]
                iri_1: str = IRIHelper_combine(value, match_value_1[0])
                this.add_reverse_mapping(term, iri_1)



    def try_find_term(self, term: str) -> str | None:
        this: LDContext = self
        definition: str | None
        match_value: str | None
        pattern_input: tuple[bool, str]
        out_arg: str = None
        def _arrow1016(__unit: None=None) -> str:
            return out_arg

        def _arrow1017(v: str) -> None:
            nonlocal out_arg
            out_arg = v

        pattern_input = (try_get_value(this.mappings_004045, term, FSharpRef(_arrow1016, _arrow1017)), out_arg)
        match_value = pattern_input[1] if pattern_input[0] else None
        def chooser(ctx: LDContext) -> str | None:
            return ctx.TryResolveTerm(term)

        definition = try_pick(chooser, this.base_contexts_004041) if (match_value is None) else match_value
        if definition is None:
            return None

        else: 
            active_pattern_result: tuple[str, str] | None = IRIHelper__007CCompactIri_007C__007C(definition)
            def _arrow1018(__unit: None=None) -> str | None:
                prefix: str = active_pattern_result[0]
                suffix: str = active_pattern_result[1]
                return IRIHelper_combine(prefix if (prefix == term) else default_arg(this.try_find_term(prefix), prefix), suffix if (suffix == term) else default_arg(this.try_find_term(suffix), suffix))

            def _arrow1019(__unit: None=None) -> str | None:
                d: str = definition
                return d

            return _arrow1018() if (active_pattern_result is not None) else _arrow1019()


    def try_find_iri(self, iri: str) -> str | None:
        this: LDContext = self
        match_value: str | None
        pattern_input: tuple[bool, str]
        out_arg: str = None
        def _arrow1020(__unit: None=None) -> str:
            return out_arg

        def _arrow1021(v: str) -> None:
            nonlocal out_arg
            out_arg = v

        pattern_input = (try_get_value(this.reverse_mappings, iri, FSharpRef(_arrow1020, _arrow1021)), out_arg)
        match_value = pattern_input[1] if pattern_input[0] else None
        def chooser(ctx: LDContext) -> str | None:
            return ctx.TryGetTerm(iri)

        return try_pick(chooser, this.base_contexts_004041) if (match_value is None) else match_value

    def try_compact_iri(self, iri: str) -> Any:
        raise Exception("TryCompactIRI is Not implemented yet")


LDContext_reflection = _expr1022

def LDContext__ctor_7878CD77(mappings: Any | None=None, base_contexts: Array[LDContext] | None=None) -> LDContext:
    return LDContext(mappings, base_contexts)


__all__ = ["IRIHelper_compactIRIRegex", "IRIHelper__007CCompactIri_007C__007C", "IRIHelper_combine", "IRIHelper_combineOptional", "LDContext_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.types import FSharpRef
from ...arctrl_core.Table.arc_table_aux import (ColumnValueRefs, ensure_cell_hash_in_value_map)
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.Table.composite_column import CompositeColumn
from ...arctrl_core.Table.composite_header import (IOType, CompositeHeader)
from ...fable_library.array_ import (skip, map, try_item)
from ...fable_library.list import (of_array, singleton as singleton_1, FSharpList, map as map_2)
from ...fable_library.map_util import (try_get_value, get_item_from_dict)
from ...fable_library.option import default_arg
from ...fable_library.range import range_big_int
from ...fable_library.seq import (to_array, delay, map as map_1, exists, to_list, append, singleton, reduce)
from ...fable_library.types import (Array, to_string)
from ...fable_library.util import (IEnumerable_1, string_hash)
from ...fs_spreadsheet.Cells.fs_cell import FsCell
from ...fs_spreadsheet.fs_column import FsColumn
from ...fs_spreadsheet.hash_codes import merge_hashes
from .composite_cell import to_string_cells as to_string_cells_1
from .composite_header import (from_string_cells, to_string_cells)

def fix_deprecated_ioheader(string_cell_col: Array[str]) -> Array[str]:
    if len(string_cell_col) == 0:
        raise Exception("Can\'t fix IOHeader Invalid column, neither header nor values given")

    values: Array[str] = skip(1, string_cell_col, None)
    match_value: IOType = IOType.of_string(string_cell_col[0])
    if match_value.tag == 4:
        return string_cell_col

    elif match_value.tag == 0:
        string_cell_col[0] = to_string(CompositeHeader(11, IOType(0)))
        return string_cell_col

    else: 
        string_cell_col[0] = to_string(CompositeHeader(12, match_value))
        return string_cell_col



def from_string_cell_columns(columns: Array[Array[str]]) -> CompositeColumn:
    def mapping(c: Array[str], columns: Any=columns) -> str:
        return c[0]

    pattern_input: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = from_string_cells(map(mapping, columns, None))
    l: int = len(columns[0]) or 0
    def _arrow3537(__unit: None=None, columns: Any=columns) -> IEnumerable_1[CompositeCell]:
        def _arrow3536(i: int) -> CompositeCell:
            def mapping_1(c_1: Array[str]) -> str:
                return c_1[i]

            return pattern_input[1](map(mapping_1, columns, None))

        return map_1(_arrow3536, range_big_int(1, 1, l - 1))

    cells: Array[CompositeCell] = list(to_array(delay(_arrow3537)))
    return CompositeColumn.create(pattern_input[0], cells)


def string_cell_columns_of_fs_columns(columns: Array[FsColumn]) -> Array[Array[str]]:
    def mapping_1(c: FsColumn, columns: Any=columns) -> Array[str]:
        c.ToDenseColumn()
        def mapping(cell: FsCell, c: Any=c) -> str:
            return cell.ValueAsString()

        return map(mapping, to_array(c.Cells), None)

    return map(mapping_1, columns, None)


def from_fs_columns(columns: Array[FsColumn]) -> CompositeColumn:
    def mapping_1(c: FsColumn, columns: Any=columns) -> Array[str]:
        c.ToDenseColumn()
        def mapping(c_1: FsCell, c: Any=c) -> str:
            return c_1.ValueAsString()

        return map(mapping, to_array(c.Cells), None)

    return from_string_cell_columns(map(mapping_1, columns, None))


def to_string_cell_columns(column: CompositeColumn) -> FSharpList[FSharpList[str]]:
    def predicate(c: CompositeCell, column: Any=column) -> bool:
        return c.is_unitized

    has_unit: bool = exists(predicate, column.Cells)
    is_term: bool = column.Header.IsTermColumn
    def predicate_1(c_1: CompositeCell, column: Any=column) -> bool:
        return c_1.is_data

    is_data: bool = exists(predicate_1, column.Cells) if column.Header.IsDataColumn else False
    header: Array[str] = to_string_cells(has_unit, column.Header)
    composite_cells: Array[CompositeCell] = to_array(column.Cells)
    def mapping(cell: CompositeCell, column: Any=column) -> Array[str]:
        return to_string_cells_1(is_term, has_unit, cell)

    cells: Array[Array[str]] = map(mapping, to_array(column.Cells), None)
    def get_cell_or_default(ri: int, ci: int, cells_1: Array[Array[str]], column: Any=column) -> str:
        return default_arg(try_item(ci, cells_1[ri]), "")

    if has_unit:
        def _arrow3543(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3542(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3541(i: int) -> str:
                    return get_cell_or_default(i, 0, cells)

                return map_1(_arrow3541, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[0]), delay(_arrow3542))

        def _arrow3546(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3545(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3544(i_1: int) -> str:
                    return get_cell_or_default(i_1, 1, cells)

                return map_1(_arrow3544, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[1]), delay(_arrow3545))

        def _arrow3549(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3548(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3547(i_2: int) -> str:
                    return get_cell_or_default(i_2, 2, cells)

                return map_1(_arrow3547, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[2]), delay(_arrow3548))

        def _arrow3552(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3551(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3550(i_3: int) -> str:
                    return get_cell_or_default(i_3, 3, cells)

                return map_1(_arrow3550, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[3]), delay(_arrow3551))

        return of_array([to_list(delay(_arrow3543)), to_list(delay(_arrow3546)), to_list(delay(_arrow3549)), to_list(delay(_arrow3552))])

    elif is_term:
        def _arrow3558(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3557(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3556(i_4: int) -> str:
                    return get_cell_or_default(i_4, 0, cells)

                return map_1(_arrow3556, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[0]), delay(_arrow3557))

        def _arrow3561(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3560(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3559(i_5: int) -> str:
                    return get_cell_or_default(i_5, 1, cells)

                return map_1(_arrow3559, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[1]), delay(_arrow3560))

        def _arrow3564(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3563(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3562(i_6: int) -> str:
                    return get_cell_or_default(i_6, 2, cells)

                return map_1(_arrow3562, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[2]), delay(_arrow3563))

        return of_array([to_list(delay(_arrow3558)), to_list(delay(_arrow3561)), to_list(delay(_arrow3564))])

    elif is_data:
        def _arrow3570(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3569(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3568(i_7: int) -> str:
                    return get_cell_or_default(i_7, 0, cells)

                return map_1(_arrow3568, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[0]), delay(_arrow3569))

        def _arrow3573(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3572(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3571(i_8: int) -> str:
                    return get_cell_or_default(i_8, 1, cells)

                return map_1(_arrow3571, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[1]), delay(_arrow3572))

        def _arrow3576(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3575(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3574(i_9: int) -> str:
                    return get_cell_or_default(i_9, 2, cells)

                return map_1(_arrow3574, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[2]), delay(_arrow3575))

        return of_array([to_list(delay(_arrow3570)), to_list(delay(_arrow3573)), to_list(delay(_arrow3576))])

    else: 
        def _arrow3579(__unit: None=None, column: Any=column) -> IEnumerable_1[str]:
            def _arrow3578(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3577(i_10: int) -> str:
                    return cells[i_10][0]

                return map_1(_arrow3577, range_big_int(0, 1, len(composite_cells) - 1))

            return append(singleton(header[0]), delay(_arrow3578))

        return singleton_1(to_list(delay(_arrow3579)))



def to_fs_columns(column: CompositeColumn) -> FSharpList[FSharpList[FsCell]]:
    def mapping_1(c: FSharpList[str], column: Any=column) -> FSharpList[FsCell]:
        def mapping(s: str, c: Any=c) -> FsCell:
            return FsCell(s)

        return map_2(mapping, c)

    return map_2(mapping_1, to_string_cell_columns(column))


def ColumnValueRefs_fromStringCellColumns(value_map: Any, columns: Array[Array[str]]) -> tuple[CompositeHeader, ColumnValueRefs]:
    hash_map: Any = dict([])
    cells: Any = dict([])
    constant: bool = True
    def mapping(c: Array[str], value_map: Any=value_map, columns: Any=columns) -> str:
        return c[0]

    pattern_input: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = from_string_cells(map(mapping, columns, None))
    header: CompositeHeader = pattern_input[0]
    cell_parser: Callable[[Array[str]], CompositeCell] = pattern_input[1]
    l: int = len(columns[0]) or 0
    if l == 1:
        return (header, ColumnValueRefs(1, cells))

    else: 
        for i in range(1, (l - 1) + 1, 1):
            def mapping_1(c_1: Array[str], value_map: Any=value_map, columns: Any=columns) -> str:
                return c_1[i]

            strings: Array[str] = map(mapping_1, columns, None)
            def _arrow3580(hash_1: int, hash_2: int, value_map: Any=value_map, columns: Any=columns) -> int:
                return merge_hashes(hash_1, hash_2)

            def mapping_2(s: str, value_map: Any=value_map, columns: Any=columns) -> int:
                return string_hash(s)

            hash_3: int = reduce(_arrow3580, map_1(mapping_2, strings)) or 0
            match_value: int | None
            pattern_input_1: tuple[bool, int]
            out_arg: int = None or 0
            def _arrow3581(__unit: None=None, value_map: Any=value_map, columns: Any=columns) -> int:
                return out_arg

            def _arrow3582(v: int, value_map: Any=value_map, columns: Any=columns) -> None:
                nonlocal out_arg
                out_arg = v or 0

            pattern_input_1 = (try_get_value(hash_map, hash_3, FSharpRef(_arrow3581, _arrow3582)), out_arg)
            match_value = pattern_input_1[1] if pattern_input_1[0] else None
            if match_value is None:
                if i == 1:
                    cell_hash_3: int = ensure_cell_hash_in_value_map(cell_parser(strings), value_map) or 0
                    hash_map[hash_3] = cell_hash_3
                    cells[i - 1] = cell_hash_3

                else: 
                    if constant:
                        constant = False
                        for j in range(0, (i - 1) + 1, 1):
                            cell_hash_4: int = get_item_from_dict(cells, 0) or 0
                            cells[j] = cell_hash_4

                    cell_hash_5: int = ensure_cell_hash_in_value_map(cell_parser(strings), value_map) or 0
                    hash_map[hash_3] = cell_hash_5
                    cells[i - 1] = cell_hash_5


            else: 
                def _arrow3583(__unit: None=None, value_map: Any=value_map, columns: Any=columns) -> bool:
                    cell_hash: int = match_value or 0
                    return constant

                if _arrow3583():
                    cell_hash_1: int = match_value or 0

                else: 
                    cell_hash_2: int = match_value or 0
                    cells[i - 1] = cell_hash_2


        if constant:
            return (header, ColumnValueRefs(0, get_item_from_dict(cells, 0)))

        else: 
            return (header, ColumnValueRefs(1, cells))




__all__ = ["fix_deprecated_ioheader", "from_string_cell_columns", "string_cell_columns_of_fs_columns", "from_fs_columns", "to_string_cell_columns", "to_fs_columns", "ColumnValueRefs_fromStringCellColumns"]




from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

from .array_ import chunk_by_size as chunk_by_size_1
from .array_ import fold_back as fold_back_1
from .array_ import fold_back2 as fold_back2_1
from .array_ import map as map_1
from .array_ import map_fold as map_fold_1
from .array_ import map_fold_back as map_fold_back_1
from .array_ import pairwise as pairwise_1
from .array_ import permute as permute_1
from .array_ import reduce_back as reduce_back_1
from .array_ import reverse as reverse_1
from .array_ import scan_back as scan_back_1
from .array_ import singleton as singleton_1
from .array_ import sort_in_place_with
from .array_ import split_into as split_into_1
from .array_ import transpose as transpose_1
from .array_ import try_find_back as try_find_back_1
from .array_ import try_find_index_back as try_find_index_back_1
from .array_ import try_head as try_head_1
from .array_ import try_item as try_item_1
from .array_ import windowed as windowed_1
from .fsharp_core import Operators_NullArg
from .global_ import IGenericAdder_1, IGenericAverager_1, SR_indexOutOfBounds
from .list import FSharpList
from .list import is_empty as is_empty_1
from .list import length as length_1
from .list import of_array as of_array_1
from .list import of_seq as of_seq_1
from .list import to_array as to_array_1
from .list import try_head as try_head_2
from .list import try_item as try_item_2
from .option import some
from .option import value as value_1
from .reflection import TypeInfo, class_type
from .types import Array, to_string
from .util import (
    IComparer_1,
    IDisposable,
    IEnumerable_1,
    IEnumerator,
    IEqualityComparer_1,
    clear,
    equals,
    get_enumerator,
    ignore,
    is_array_like,
    is_disposable,
    lock,
    to_enumerable,
    to_iterator,
)
from .util import dispose as dispose_2


_T = TypeVar("_T")

_U = TypeVar("_U")

_STATE = TypeVar("_STATE")

__B = TypeVar("__B")

__A = TypeVar("__A")

_T1 = TypeVar("_T1")

_T2 = TypeVar("_T2")

_T3 = TypeVar("_T3")

_RESULT = TypeVar("_RESULT")

_COLLECTION = TypeVar("_COLLECTION")

SR_enumerationAlreadyFinished: str = "Enumeration already finished."

SR_enumerationNotStarted: str = "Enumeration has not started. Call MoveNext."

SR_inputSequenceEmpty: str = "The input sequence was empty."

SR_inputSequenceTooLong: str = "The input sequence contains more than one element."

SR_keyNotFoundAlt: str = "An index satisfying the predicate was not found in the collection."

SR_notEnoughElements: str = "The input sequence has an insufficient number of elements."

SR_resetNotSupported: str = "Reset is not supported on this enumerator."


def Enumerator_noReset(__unit: None = None) -> Any:
    raise Exception(SR_resetNotSupported)


def Enumerator_notStarted(__unit: None = None) -> Any:
    raise Exception(SR_enumerationNotStarted)


def Enumerator_alreadyFinished(__unit: None = None) -> Any:
    raise Exception(SR_enumerationAlreadyFinished)


def _expr134(gen0: TypeInfo) -> TypeInfo:
    return class_type("SeqModule.Enumerator.Seq", [gen0], Enumerator_Seq)


class Enumerator_Seq(Generic[_T]):
    def __init__(self, f: Callable[[], IEnumerator[_T]]) -> None:
        self.f: Callable[[], IEnumerator[_T]] = f

    def __str__(self, __unit: None = None) -> str:
        xs: Enumerator_Seq[_T] = self
        i: int = 0
        str_1: str = "seq ["
        with get_enumerator(xs) as e:
            while e.System_Collections_IEnumerator_MoveNext() if (i < 4) else False:
                if i > 0:
                    str_1 = str_1 + "; "

                str_1 = str_1 + to_string(e.System_Collections_Generic_IEnumerator_1_get_Current())
                i = (i + 1) or 0
            if i == 4:
                str_1 = str_1 + "; ..."

            return str_1 + "]"

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        x: Enumerator_Seq[_T] = self
        return x.f(None)

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        x: Enumerator_Seq[_T] = self
        return x.f(None)


Enumerator_Seq_reflection = _expr134


def Enumerator_Seq__ctor_673A07F2(f: Callable[[], IEnumerator[_T]]) -> Enumerator_Seq[_T]:
    return Enumerator_Seq(f)


def _expr135(gen0: TypeInfo) -> TypeInfo:
    return class_type("SeqModule.Enumerator.FromFunctions`1", [gen0], Enumerator_FromFunctions_1)


class Enumerator_FromFunctions_1(IDisposable, Generic[_T]):
    def __init__(self, current: Callable[[], _T], next_1: Callable[[], bool], dispose: Callable[[], None]) -> None:
        self.current: Callable[[], _T] = current
        self.next: Callable[[], bool] = next_1
        self.dispose: Callable[[], None] = dispose

    def System_Collections_Generic_IEnumerator_1_get_Current(self, __unit: None = None) -> _T:
        _: Enumerator_FromFunctions_1[_T] = self
        return _.current(None)

    def System_Collections_IEnumerator_get_Current(self, __unit: None = None) -> Any:
        _: Enumerator_FromFunctions_1[_T] = self
        return _.current(None)

    def System_Collections_IEnumerator_MoveNext(self, __unit: None = None) -> bool:
        _: Enumerator_FromFunctions_1[_T] = self
        return _.next(None)

    def System_Collections_IEnumerator_Reset(self, __unit: None = None) -> None:
        Enumerator_noReset()

    def Dispose(self, __unit: None = None) -> None:
        _: Enumerator_FromFunctions_1[_T] = self
        _.dispose(None)


Enumerator_FromFunctions_1_reflection = _expr135


def Enumerator_FromFunctions_1__ctor_58C54629(
    current: Callable[[], _T], next_1: Callable[[], bool], dispose: Callable[[], None]
) -> Enumerator_FromFunctions_1[_T]:
    return Enumerator_FromFunctions_1(current, next_1, dispose)


def Enumerator_cast(e: IEnumerator[Any]) -> IEnumerator[Any]:
    def current(__unit: None = None, e: Any = e) -> _T:
        return e.System_Collections_Generic_IEnumerator_1_get_Current()

    def next_1(__unit: None = None, e: Any = e) -> bool:
        return e.System_Collections_IEnumerator_MoveNext()

    def dispose(__unit: None = None, e: Any = e) -> None:
        dispose_2(e)

    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_concat(sources: IEnumerable_1[Any]) -> IEnumerator[Any]:
    outer_opt: IEnumerator[_U] | None = None
    inner_opt: IEnumerator[_T] | None = None
    started: bool = False
    finished: bool = False
    curr: _T | None = None

    def finish(__unit: None = None, sources: Any = sources) -> None:
        nonlocal finished, inner_opt, outer_opt
        finished = True
        if inner_opt is not None:
            inner: IEnumerator[_T] = inner_opt
            try:
                dispose_2(inner)

            finally:
                inner_opt = None

        if outer_opt is not None:
            outer: IEnumerator[_U] = outer_opt
            try:
                dispose_2(outer)

            finally:
                outer_opt = None

    def current(__unit: None = None, sources: Any = sources) -> _T:
        if not started:
            Enumerator_notStarted()

        elif finished:
            Enumerator_alreadyFinished()

        if curr is not None:
            return value_1(curr)

        else:
            return Enumerator_alreadyFinished()

    def next_1(__unit: None = None, sources: Any = sources) -> bool:
        nonlocal started
        if not started:
            started = True

        if finished:
            return False

        else:
            res: bool | None = None
            while res is None:
                nonlocal curr, inner_opt, outer_opt
                outer_opt_1: IEnumerator[_U] | None = outer_opt
                inner_opt_1: IEnumerator[_T] | None = inner_opt
                if outer_opt_1 is not None:
                    if inner_opt_1 is not None:
                        inner_1: IEnumerator[_T] = inner_opt_1
                        if inner_1.System_Collections_IEnumerator_MoveNext():
                            curr = some(inner_1.System_Collections_Generic_IEnumerator_1_get_Current())
                            res = True

                        else:
                            try:
                                dispose_2(inner_1)

                            finally:
                                inner_opt = None

                    else:
                        outer_1: IEnumerator[_U] = outer_opt_1
                        if outer_1.System_Collections_IEnumerator_MoveNext():
                            ie: _U = outer_1.System_Collections_Generic_IEnumerator_1_get_Current()

                            def _arrow136(__unit: None = None) -> IEnumerator[_T]:
                                copy_of_struct: _U = ie
                                return get_enumerator(copy_of_struct)

                            inner_opt = _arrow136()

                        else:
                            finish(None)
                            res = False

                else:
                    outer_opt = get_enumerator(sources)

            return value_1(res)

    def dispose(__unit: None = None, sources: Any = sources) -> None:
        if not finished:
            finish(None)

    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_enumerateThenFinally(f: Callable[[], None], e: IEnumerator[Any]) -> IEnumerator[Any]:
    def current(__unit: None = None, f: Any = f, e: Any = e) -> _T:
        return e.System_Collections_Generic_IEnumerator_1_get_Current()

    def next_1(__unit: None = None, f: Any = f, e: Any = e) -> bool:
        return e.System_Collections_IEnumerator_MoveNext()

    def dispose(__unit: None = None, f: Any = f, e: Any = e) -> None:
        try:
            dispose_2(e)

        finally:
            f(None)

    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_generateWhileSome(
    openf: Callable[[], _T], compute: Callable[[_T], _U | None], closef: Callable[[_T], None]
) -> IEnumerator[Any]:
    started: bool = False
    curr: _U | None = None
    state: _T | None = some(openf(None))

    def dispose(__unit: None = None, openf: Any = openf, compute: Any = compute, closef: Any = closef) -> None:
        nonlocal state
        if state is not None:
            x_1: _T = value_1(state)
            try:
                closef(x_1)

            finally:
                state = None

    def finish(__unit: None = None, openf: Any = openf, compute: Any = compute, closef: Any = closef) -> None:
        nonlocal curr
        try:
            dispose(None)

        finally:
            curr = None

    def current(__unit: None = None, openf: Any = openf, compute: Any = compute, closef: Any = closef) -> _U:
        if not started:
            Enumerator_notStarted()

        if curr is not None:
            return value_1(curr)

        else:
            return Enumerator_alreadyFinished()

    def next_1(__unit: None = None, openf: Any = openf, compute: Any = compute, closef: Any = closef) -> bool:
        nonlocal started, curr
        if not started:
            started = True

        if state is not None:
            s: _T = value_1(state)
            match_value_1: _U | None
            try:
                match_value_1 = compute(s)

            except Exception as match_value:
                finish(None)
                raise match_value

            if match_value_1 is not None:
                curr = match_value_1
                return True

            else:
                finish(None)
                return False

        else:
            return False

    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_unfold(f: Callable[[_STATE], tuple[_T, _STATE] | None], state: Any) -> IEnumerator[Any]:
    curr: tuple[_T, _STATE] | None = None
    acc: _STATE = state

    def current(__unit: None = None, f: Any = f, state: Any = state) -> _T:
        if curr is not None:
            x: _T = curr[0]
            st: _STATE = curr[1]
            return x

        else:
            return Enumerator_notStarted()

    def next_1(__unit: None = None, f: Any = f, state: Any = state) -> bool:
        nonlocal curr, acc
        curr = f(acc)
        if curr is not None:
            x_1: _T = curr[0]
            st_1: _STATE = curr[1]
            acc = st_1
            return True

        else:
            return False

    def dispose(__unit: None = None, f: Any = f, state: Any = state) -> None:
        pass

    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def index_not_found(__unit: None = None) -> Any:
    raise Exception(SR_keyNotFoundAlt)


def check_non_null(arg_name: str, arg: Any) -> None:
    if arg is None:
        Operators_NullArg(arg_name)


def mk_seq(f: Callable[[], IEnumerator[_T]]) -> IEnumerable_1[Any]:
    return Enumerator_Seq__ctor_673A07F2(f)


def of_seq(xs: IEnumerable_1[Any]) -> IEnumerator[Any]:
    check_non_null("source", xs)
    return get_enumerator(xs)


def delay(generator: Callable[[], IEnumerable_1[_T]]) -> IEnumerable_1[Any]:
    def _arrow137(__unit: None = None, generator: Any = generator) -> IEnumerator[_T]:
        return get_enumerator(generator(None))

    return mk_seq(_arrow137)


def concat(sources: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow138(__unit: None = None, sources: Any = sources) -> IEnumerator[_T]:
        return Enumerator_concat(sources)

    return mk_seq(_arrow138)


def unfold(generator: Callable[[_STATE], tuple[_T, _STATE] | None], state: Any) -> IEnumerable_1[Any]:
    def _arrow139(__unit: None = None, generator: Any = generator, state: Any = state) -> IEnumerator[_T]:
        return Enumerator_unfold(generator, state)

    return mk_seq(_arrow139)


def empty(__unit: None = None) -> IEnumerable_1[Any]:
    def _arrow140(__unit: None = None) -> IEnumerable_1[_T]:
        return [0] * 0

    return delay(_arrow140)


def singleton(x: Any | None = None) -> IEnumerable_1[Any]:
    def _arrow141(__unit: None = None, x: Any = x) -> IEnumerable_1[_T]:
        return singleton_1(x, None)

    return delay(_arrow141)


def of_array(arr: Array[Any]) -> IEnumerable_1[Any]:
    return arr


def to_array(xs: IEnumerable_1[Any]) -> Array[Any]:
    if isinstance(xs, FSharpList):
        return to_array_1(xs)

    else:
        return list(xs)


def of_list(xs: FSharpList[Any]) -> IEnumerable_1[Any]:
    return xs


def to_list(xs: IEnumerable_1[Any]) -> FSharpList[Any]:
    if is_array_like(xs):
        return of_array_1(xs)

    elif isinstance(xs, FSharpList):
        return xs

    else:
        return of_seq_1(xs)


def generate(
    create: Callable[[], __A], compute: Callable[[__A], __B | None], dispose: Callable[[__A], None]
) -> IEnumerable_1[Any]:
    def _arrow142(
        __unit: None = None, create: Any = create, compute: Any = compute, dispose: Any = dispose
    ) -> IEnumerator[__B]:
        return Enumerator_generateWhileSome(create, compute, dispose)

    return mk_seq(_arrow142)


def generate_indexed(
    create: Callable[[], __A], compute: Callable[[int, __A], __B | None], dispose: Callable[[__A], None]
) -> IEnumerable_1[Any]:
    def _arrow144(
        __unit: None = None, create: Any = create, compute: Any = compute, dispose: Any = dispose
    ) -> IEnumerator[__B]:
        i: int = -1

        def _arrow143(x: __A | None = None) -> __B | None:
            nonlocal i
            i = (i + 1) or 0
            return compute(i, x)

        return Enumerator_generateWhileSome(create, _arrow143, dispose)

    return mk_seq(_arrow144)


def append(xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    return concat(to_enumerable([xs, ys]))


def cast(xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow145(__unit: None = None, xs: Any = xs) -> IEnumerator[_T]:
        check_non_null("source", xs)
        return Enumerator_cast(get_enumerator(xs))

    return mk_seq(_arrow145)


def choose(chooser: Callable[[_T], _U | None], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow146(__unit: None = None, chooser: Any = chooser, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow147(e: IEnumerator[_T], chooser: Any = chooser, xs: Any = xs) -> _U | None:
        curr: _U | None = None
        while e.System_Collections_IEnumerator_MoveNext() if (curr is None) else False:
            curr = chooser(e.System_Collections_Generic_IEnumerator_1_get_Current())
        return curr

    def _arrow148(e_1: IEnumerator[_T], chooser: Any = chooser, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate(_arrow146, _arrow147, _arrow148)


def compare_with(comparer: Callable[[_T, _T], int], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> int:
    with of_seq(xs) as e1:
        with of_seq(ys) as e2:
            c: int = 0
            b1: bool = e1.System_Collections_IEnumerator_MoveNext()
            b2: bool = e2.System_Collections_IEnumerator_MoveNext()
            while b2 if (b1 if (c == 0) else False) else False:
                c = (
                    comparer(
                        e1.System_Collections_Generic_IEnumerator_1_get_Current(),
                        e2.System_Collections_Generic_IEnumerator_1_get_Current(),
                    )
                    or 0
                )
                if c == 0:
                    b1 = e1.System_Collections_IEnumerator_MoveNext()
                    b2 = e2.System_Collections_IEnumerator_MoveNext()

            if c != 0:
                return c

            elif b1:
                return 1

            elif b2:
                return -1

            else:
                return 0


def contains(value: Any, xs: IEnumerable_1[Any], comparer: IEqualityComparer_1[Any]) -> bool:
    with of_seq(xs) as e:
        found: bool = False
        while e.System_Collections_IEnumerator_MoveNext() if (not found) else False:
            found = comparer.Equals(value, e.System_Collections_Generic_IEnumerator_1_get_Current())
        return found


def enumerate_from_functions(
    create: Callable[[], __A], move_next: Callable[[__A], bool], current: Callable[[__A], __B]
) -> IEnumerable_1[Any]:
    def _arrow149(
        x: __A | None = None, create: Any = create, move_next: Any = move_next, current: Any = current
    ) -> __B | None:
        return some(current(x)) if move_next(x) else None

    def _arrow150(
        x_1: __A | None = None, create: Any = create, move_next: Any = move_next, current: Any = current
    ) -> None:
        match_value: Any = x_1
        if is_disposable(match_value):
            dispose_2(match_value)

    return generate(create, _arrow149, _arrow150)


def enumerate_then_finally(source: IEnumerable_1[Any], compensation: Callable[[], None]) -> IEnumerable_1[Any]:
    compensation_1: Callable[[], None] = compensation

    def _arrow151(__unit: None = None, source: Any = source, compensation: Any = compensation) -> IEnumerator[_T]:
        try:
            return Enumerator_enumerateThenFinally(compensation_1, of_seq(source))

        except Exception as match_value:
            compensation_1(None)
            raise match_value

    return mk_seq(_arrow151)


def enumerate_using(resource: Any, source: Callable[[_T], __A]) -> IEnumerable_1[Any]:
    def compensation(__unit: None = None, resource: Any = resource, source: Any = source) -> None:
        if equals(resource, None):
            pass

        else:
            copy_of_struct: _T = resource
            dispose_2(copy_of_struct)

    def _arrow152(__unit: None = None, resource: Any = resource, source: Any = source) -> IEnumerator[_U]:
        try:
            return Enumerator_enumerateThenFinally(compensation, of_seq(source(resource)))

        except Exception as match_value_1:
            compensation(None)
            raise match_value_1

    return mk_seq(_arrow152)


def enumerate_while(guard: Callable[[], bool], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow153(i: int, guard: Any = guard, xs: Any = xs) -> tuple[IEnumerable_1[_T], int] | None:
        return ((xs, i + 1)) if guard(None) else None

    return concat(unfold(_arrow153, 0))


def filter(f: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def chooser(x: _T | None = None, f: Any = f, xs: Any = xs) -> _T | None:
        if f(x):
            return some(x)

        else:
            return None

    return choose(chooser, xs)


def exists(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> bool:
    with of_seq(xs) as e:
        found: bool = False
        while e.System_Collections_IEnumerator_MoveNext() if (not found) else False:
            found = predicate(e.System_Collections_Generic_IEnumerator_1_get_Current())
        return found


def exists2(predicate: Callable[[_T1, _T2], bool], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> bool:
    with of_seq(xs) as e1:
        with of_seq(ys) as e2:
            found: bool = False
            while (
                e2.System_Collections_IEnumerator_MoveNext()
                if (e1.System_Collections_IEnumerator_MoveNext() if (not found) else False)
                else False
            ):
                found = predicate(
                    e1.System_Collections_Generic_IEnumerator_1_get_Current(),
                    e2.System_Collections_Generic_IEnumerator_1_get_Current(),
                )
            return found


def exactly_one(xs: IEnumerable_1[Any]) -> Any:
    with of_seq(xs) as e:
        if e.System_Collections_IEnumerator_MoveNext():
            v: _T = e.System_Collections_Generic_IEnumerator_1_get_Current()
            if e.System_Collections_IEnumerator_MoveNext():
                raise Exception((SR_inputSequenceTooLong + "\\nParameter name: ") + "source")

            else:
                return v

        else:
            raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")


def try_exactly_one(xs: IEnumerable_1[Any]) -> Any | None:
    with of_seq(xs) as e:
        if e.System_Collections_IEnumerator_MoveNext():
            v: _T = e.System_Collections_Generic_IEnumerator_1_get_Current()
            if e.System_Collections_IEnumerator_MoveNext():
                return None

            else:
                return some(v)

        else:
            return None


def try_find(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> Any | None:
    with of_seq(xs) as e:
        res: _T | None = None
        while e.System_Collections_IEnumerator_MoveNext() if (res is None) else False:
            c: _T = e.System_Collections_Generic_IEnumerator_1_get_Current()
            if predicate(c):
                res = some(c)

        return res


def find(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> Any:
    match_value: _T | None = try_find(predicate, xs)
    if match_value is None:
        return index_not_found()

    else:
        return value_1(match_value)


def try_find_back(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> Any | None:
    return try_find_back_1(predicate, to_array(xs))


def find_back(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> Any:
    match_value: _T | None = try_find_back(predicate, xs)
    if match_value is None:
        return index_not_found()

    else:
        return value_1(match_value)


def try_find_index(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> int | None:
    with of_seq(xs) as e:

        def loop(i_mut: int, predicate: Any = predicate, xs: Any = xs) -> int | None:
            while True:
                (i,) = (i_mut,)
                if e.System_Collections_IEnumerator_MoveNext():
                    if predicate(e.System_Collections_Generic_IEnumerator_1_get_Current()):
                        return i

                    else:
                        i_mut = i + 1
                        continue

                else:
                    return None

                break

        return loop(0)


def find_index(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> int:
    match_value: int | None = try_find_index(predicate, xs)
    if match_value is None:
        index_not_found()
        return -1

    else:
        return match_value


def try_find_index_back(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> int | None:
    return try_find_index_back_1(predicate, to_array(xs))


def find_index_back(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> int:
    match_value: int | None = try_find_index_back(predicate, xs)
    if match_value is None:
        index_not_found()
        return -1

    else:
        return match_value


def fold(folder: Callable[[_STATE, _T], _STATE], state: Any, xs: IEnumerable_1[Any]) -> Any:
    with of_seq(xs) as e:
        acc: _STATE = state
        while e.System_Collections_IEnumerator_MoveNext():
            acc = folder(acc, e.System_Collections_Generic_IEnumerator_1_get_Current())
        return acc


def fold_back(folder: Callable[[_T, Any], Any], xs: IEnumerable_1[Any], state: Any = None) -> Any:
    return fold_back_1(folder, to_array(xs), state)


def fold2(
    folder: Callable[[_STATE, _T1, _T2], _STATE], state: Any, xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]
) -> Any:
    with of_seq(xs) as e1:
        with of_seq(ys) as e2:
            acc: _STATE = state
            while (
                e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False
            ):
                acc = folder(
                    acc,
                    e1.System_Collections_Generic_IEnumerator_1_get_Current(),
                    e2.System_Collections_Generic_IEnumerator_1_get_Current(),
                )
            return acc


def fold_back2(
    folder: Callable[[_T1, _T2, _STATE], _STATE], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any], state: Any
) -> Any:
    return fold_back2_1(folder, to_array(xs), to_array(ys), state)


def for_all(predicate: Callable[[__A], bool], xs: IEnumerable_1[Any]) -> bool:
    def _arrow154(x: __A | None = None, predicate: Any = predicate, xs: Any = xs) -> bool:
        return not predicate(x)

    return not exists(_arrow154, xs)


def for_all2(predicate: Callable[[__A, __B], bool], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> bool:
    def _arrow155(x: __A, y: __B, predicate: Any = predicate, xs: Any = xs, ys: Any = ys) -> bool:
        return not predicate(x, y)

    return not exists2(_arrow155, xs, ys)


def try_head(xs: IEnumerable_1[Any]) -> Any | None:
    if is_array_like(xs):
        return try_head_1(xs)

    elif isinstance(xs, FSharpList):
        return try_head_2(xs)

    else:
        with of_seq(xs) as e:
            if e.System_Collections_IEnumerator_MoveNext():
                return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

            else:
                return None


def head(xs: IEnumerable_1[Any]) -> Any:
    match_value: _T | None = try_head(xs)
    if match_value is None:
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")

    else:
        return value_1(match_value)


def initialize(count: int, f: Callable[[int], __A]) -> IEnumerable_1[Any]:
    def _arrow156(i: int, count: Any = count, f: Any = f) -> tuple[__A, int] | None:
        return ((f(i), i + 1)) if (i < count) else None

    return unfold(_arrow156, 0)


def initialize_infinite(f: Callable[[int], __A]) -> IEnumerable_1[Any]:
    return initialize(2147483647, f)


def is_empty(xs: IEnumerable_1[Any]) -> bool:
    if is_array_like(xs):
        return len(xs) == 0

    elif isinstance(xs, FSharpList):
        return is_empty_1(xs)

    else:
        with of_seq(xs) as e:
            return not e.System_Collections_IEnumerator_MoveNext()


def try_item(index: int, xs: IEnumerable_1[Any]) -> Any | None:
    if is_array_like(xs):
        return try_item_1(index, xs)

    elif isinstance(xs, FSharpList):
        return try_item_2(index, xs)

    else:
        with of_seq(xs) as e:

            def loop(index_1_mut: int, index: Any = index, xs: Any = xs) -> _T | None:
                while True:
                    (index_1,) = (index_1_mut,)
                    if not e.System_Collections_IEnumerator_MoveNext():
                        return None

                    elif index_1 == 0:
                        return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

                    else:
                        index_1_mut = index_1 - 1
                        continue

                    break

            return loop(index)


def item(index: int, xs: IEnumerable_1[Any]) -> Any:
    match_value: _T | None = try_item(index, xs)
    if match_value is None:
        raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "index")

    else:
        return value_1(match_value)


def iterate(action: Callable[[__A], None], xs: IEnumerable_1[Any]) -> None:
    def _arrow157(unit_var: None, x: __A, action: Any = action, xs: Any = xs) -> None:
        action(x)

    fold(_arrow157, None, xs)


def iterate2(action: Callable[[__A, __B], None], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> None:
    def _arrow158(unit_var: None, x: __A, y: __B, action: Any = action, xs: Any = xs, ys: Any = ys) -> None:
        action(x, y)

    fold2(_arrow158, None, xs, ys)


def iterate_indexed(action: Callable[[int, __A], None], xs: IEnumerable_1[Any]) -> None:
    def _arrow159(i: int, x: __A, action: Any = action, xs: Any = xs) -> int:
        action(i, x)
        return i + 1

    ignore(fold(_arrow159, 0, xs))


def iterate_indexed2(action: Callable[[int, __A, __B], None], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> None:
    def _arrow160(i: int, x: __A, y: __B, action: Any = action, xs: Any = xs, ys: Any = ys) -> int:
        action(i, x, y)
        return i + 1

    ignore(fold2(_arrow160, 0, xs, ys))


def try_last(xs: IEnumerable_1[Any]) -> Any | None:
    with of_seq(xs) as e:

        def loop(acc_mut: _T | None = None, xs: Any = xs) -> _T | None:
            while True:
                (acc,) = (acc_mut,)
                if not e.System_Collections_IEnumerator_MoveNext():
                    return acc

                else:
                    acc_mut = e.System_Collections_Generic_IEnumerator_1_get_Current()
                    continue

                break

        if e.System_Collections_IEnumerator_MoveNext():
            return some(loop(e.System_Collections_Generic_IEnumerator_1_get_Current()))

        else:
            return None


def last(xs: IEnumerable_1[Any]) -> Any:
    match_value: _T | None = try_last(xs)
    if match_value is None:
        raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "source")

    else:
        return value_1(match_value)


def length(xs: IEnumerable_1[Any]) -> int:
    if is_array_like(xs):
        return len(xs)

    elif isinstance(xs, FSharpList):
        return length_1(xs)

    else:
        with of_seq(xs) as e:
            count: int = 0
            while e.System_Collections_IEnumerator_MoveNext():
                count = (count + 1) or 0
            return count


def map(mapping: Callable[[_T], _U], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow161(__unit: None = None, mapping: Any = mapping, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow162(e: IEnumerator[_T], mapping: Any = mapping, xs: Any = xs) -> _U | None:
        return (
            some(mapping(e.System_Collections_Generic_IEnumerator_1_get_Current()))
            if e.System_Collections_IEnumerator_MoveNext()
            else None
        )

    def _arrow163(e_1: IEnumerator[_T], mapping: Any = mapping, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate(_arrow161, _arrow162, _arrow163)


def map_indexed(mapping: Callable[[int, _T], _U], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow164(__unit: None = None, mapping: Any = mapping, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow165(i: int, e: IEnumerator[_T], mapping: Any = mapping, xs: Any = xs) -> _U | None:
        return (
            some(mapping(i, e.System_Collections_Generic_IEnumerator_1_get_Current()))
            if e.System_Collections_IEnumerator_MoveNext()
            else None
        )

    def _arrow166(e_1: IEnumerator[_T], mapping: Any = mapping, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow164, _arrow165, _arrow166)


def indexed(xs: IEnumerable_1[Any]) -> IEnumerable_1[tuple[int, _T]]:
    def mapping(i: int, x: _T, xs: Any = xs) -> tuple[int, _T]:
        return (i, x)

    return map_indexed(mapping, xs)


def map2(mapping: Callable[[_T1, _T2], _U], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow167(
        __unit: None = None, mapping: Any = mapping, xs: Any = xs, ys: Any = ys
    ) -> tuple[IEnumerator[_T1], IEnumerator[_T2]]:
        return (of_seq(xs), of_seq(ys))

    def _arrow168(
        tupled_arg: tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Any = mapping, xs: Any = xs, ys: Any = ys
    ) -> _U | None:
        e1: IEnumerator[_T1] = tupled_arg[0]
        e2: IEnumerator[_T2] = tupled_arg[1]
        return (
            some(
                mapping(
                    e1.System_Collections_Generic_IEnumerator_1_get_Current(),
                    e2.System_Collections_Generic_IEnumerator_1_get_Current(),
                )
            )
            if (e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False)
            else None
        )

    def _arrow169(
        tupled_arg_1: tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Any = mapping, xs: Any = xs, ys: Any = ys
    ) -> None:
        try:
            dispose_2(tupled_arg_1[0])

        finally:
            dispose_2(tupled_arg_1[1])

    return generate(_arrow167, _arrow168, _arrow169)


def map_indexed2(
    mapping: Callable[[int, _T1, _T2], _U], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]
) -> IEnumerable_1[Any]:
    def _arrow170(
        __unit: None = None, mapping: Any = mapping, xs: Any = xs, ys: Any = ys
    ) -> tuple[IEnumerator[_T1], IEnumerator[_T2]]:
        return (of_seq(xs), of_seq(ys))

    def _arrow171(
        i: int,
        tupled_arg: tuple[IEnumerator[_T1], IEnumerator[_T2]],
        mapping: Any = mapping,
        xs: Any = xs,
        ys: Any = ys,
    ) -> _U | None:
        e1: IEnumerator[_T1] = tupled_arg[0]
        e2: IEnumerator[_T2] = tupled_arg[1]
        return (
            some(
                mapping(
                    i,
                    e1.System_Collections_Generic_IEnumerator_1_get_Current(),
                    e2.System_Collections_Generic_IEnumerator_1_get_Current(),
                )
            )
            if (e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False)
            else None
        )

    def _arrow172(
        tupled_arg_1: tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Any = mapping, xs: Any = xs, ys: Any = ys
    ) -> None:
        try:
            dispose_2(tupled_arg_1[0])

        finally:
            dispose_2(tupled_arg_1[1])

    return generate_indexed(_arrow170, _arrow171, _arrow172)


def map3(
    mapping: Callable[[_T1, _T2, _T3], _U], xs: IEnumerable_1[Any], ys: IEnumerable_1[Any], zs: IEnumerable_1[Any]
) -> IEnumerable_1[Any]:
    def _arrow173(
        __unit: None = None, mapping: Any = mapping, xs: Any = xs, ys: Any = ys, zs: Any = zs
    ) -> tuple[IEnumerator[_T1], IEnumerator[_T2], IEnumerator[_T3]]:
        return (of_seq(xs), of_seq(ys), of_seq(zs))

    def _arrow174(
        tupled_arg: tuple[IEnumerator[_T1], IEnumerator[_T2], IEnumerator[_T3]],
        mapping: Any = mapping,
        xs: Any = xs,
        ys: Any = ys,
        zs: Any = zs,
    ) -> _U | None:
        e1: IEnumerator[_T1] = tupled_arg[0]
        e2: IEnumerator[_T2] = tupled_arg[1]
        e3: IEnumerator[_T3] = tupled_arg[2]
        return (
            some(
                mapping(
                    e1.System_Collections_Generic_IEnumerator_1_get_Current(),
                    e2.System_Collections_Generic_IEnumerator_1_get_Current(),
                    e3.System_Collections_Generic_IEnumerator_1_get_Current(),
                )
            )
            if (
                e3.System_Collections_IEnumerator_MoveNext()
                if (
                    e2.System_Collections_IEnumerator_MoveNext()
                    if e1.System_Collections_IEnumerator_MoveNext()
                    else False
                )
                else False
            )
            else None
        )

    def _arrow175(
        tupled_arg_1: tuple[IEnumerator[_T1], IEnumerator[_T2], IEnumerator[_T3]],
        mapping: Any = mapping,
        xs: Any = xs,
        ys: Any = ys,
        zs: Any = zs,
    ) -> None:
        try:
            dispose_2(tupled_arg_1[0])

        finally:
            try:
                dispose_2(tupled_arg_1[1])

            finally:
                dispose_2(tupled_arg_1[2])

    return generate(_arrow173, _arrow174, _arrow175)


def read_only(xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    check_non_null("source", xs)

    def _arrow176(x: _T | None = None, xs: Any = xs) -> _T | None:
        return x

    return map(_arrow176, xs)


def _expr177(gen0: TypeInfo) -> TypeInfo:
    return class_type("SeqModule.CachedSeq`1", [gen0], CachedSeq_1)


class CachedSeq_1(IDisposable, Generic[_T]):
    def __init__(self, cleanup: Callable[[], None], res: IEnumerable_1[Any]) -> None:
        self.cleanup: Callable[[], None] = cleanup
        self.res: IEnumerable_1[_T] = res

    def Dispose(self, __unit: None = None) -> None:
        _: CachedSeq_1[_T] = self
        _.cleanup(None)

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        _: CachedSeq_1[_T] = self
        return get_enumerator(_.res)

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        _: CachedSeq_1[_T] = self
        return get_enumerator(_.res)


CachedSeq_1_reflection = _expr177


def CachedSeq_1__ctor_Z7A8347D4(cleanup: Callable[[], None], res: IEnumerable_1[Any]) -> CachedSeq_1[_T]:
    return CachedSeq_1(cleanup, res)


def CachedSeq_1__Clear(_: CachedSeq_1[Any]) -> None:
    _.cleanup(None)


def cache(source: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    check_non_null("source", source)
    prefix: Array[_T] = []
    enumerator_r: (IEnumerator[_T] | None) | None = None

    def cleanup(__unit: None = None, source: Any = source) -> None:
        def action_1(__unit: None = None) -> None:
            nonlocal enumerator_r
            clear(prefix)
            (pattern_matching_result, e) = (None, None)
            if enumerator_r is not None:
                if value_1(enumerator_r) is not None:
                    pattern_matching_result = 0
                    e = value_1(enumerator_r)

                else:
                    pattern_matching_result = 1

            else:
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                dispose_2(e)

            enumerator_r = None

        lock(prefix, action_1)

    def _arrow178(i_1: int, source: Any = source) -> tuple[_T, int] | None:
        def action(__unit: None = None) -> tuple[_T, int] | None:
            nonlocal enumerator_r
            if i_1 < len(prefix):
                return (prefix[i_1], i_1 + 1)

            else:
                if i_1 >= len(prefix):
                    opt_enumerator_2: IEnumerator[_T] | None
                    if enumerator_r is not None:
                        opt_enumerator_2 = value_1(enumerator_r)

                    else:
                        opt_enumerator: IEnumerator[_T] | None = get_enumerator(source)
                        enumerator_r = some(opt_enumerator)
                        opt_enumerator_2 = opt_enumerator

                    if opt_enumerator_2 is None:
                        pass

                    else:
                        enumerator: IEnumerator[_T] = opt_enumerator_2
                        if enumerator.System_Collections_IEnumerator_MoveNext():
                            (prefix.append(enumerator.System_Collections_Generic_IEnumerator_1_get_Current()))

                        else:
                            dispose_2(enumerator)
                            enumerator_r = some(None)

                if i_1 < len(prefix):
                    return (prefix[i_1], i_1 + 1)

                else:
                    return None

        return lock(prefix, action)

    return CachedSeq_1__ctor_Z7A8347D4(cleanup, unfold(_arrow178, 0))


def all_pairs(xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> IEnumerable_1[tuple[_T1, _T2]]:
    ys_cache: IEnumerable_1[_T2] = cache(ys)

    def _arrow179(__unit: None = None, xs: Any = xs, ys: Any = ys) -> IEnumerable_1[tuple[_T1, _T2]]:
        def mapping_1(x: _T1 | None = None) -> IEnumerable_1[tuple[_T1, _T2]]:
            def mapping(y: _T2 | None = None, x: Any = x) -> tuple[_T1, _T2]:
                return (x, y)

            return map(mapping, ys_cache)

        return concat(map(mapping_1, xs))

    return delay(_arrow179)


def map_fold(
    mapping: Callable[[_STATE, _T], tuple[_RESULT, _STATE]], state: Any, xs: IEnumerable_1[Any]
) -> tuple[IEnumerable_1[_RESULT], _STATE]:
    pattern_input: tuple[Array[_RESULT], _STATE] = map_fold_1(mapping, state, to_array(xs), None)
    return (read_only(pattern_input[0]), pattern_input[1])


def map_fold_back(
    mapping: Callable[[_T, _STATE], tuple[_RESULT, _STATE]], xs: IEnumerable_1[Any], state: Any
) -> tuple[IEnumerable_1[_RESULT], _STATE]:
    pattern_input: tuple[Array[_RESULT], _STATE] = map_fold_back_1(mapping, to_array(xs), state, None)
    return (read_only(pattern_input[0]), pattern_input[1])


def try_pick(chooser: Callable[[_T], __A | None], xs: IEnumerable_1[Any]) -> Any | None:
    with of_seq(xs) as e:
        res: __A | None = None
        while e.System_Collections_IEnumerator_MoveNext() if (res is None) else False:
            res = chooser(e.System_Collections_Generic_IEnumerator_1_get_Current())
        return res


def pick(chooser: Callable[[_T], __A | None], xs: IEnumerable_1[Any]) -> Any:
    match_value: __A | None = try_pick(chooser, xs)
    if match_value is None:
        return index_not_found()

    else:
        return value_1(match_value)


def reduce(folder: Callable[[_T, _T], _T], xs: IEnumerable_1[Any]) -> Any:
    with of_seq(xs) as e:

        def loop(acc_mut: _T | None = None, folder: Any = folder, xs: Any = xs) -> _T | None:
            while True:
                (acc,) = (acc_mut,)
                if e.System_Collections_IEnumerator_MoveNext():
                    acc_mut = folder(acc, e.System_Collections_Generic_IEnumerator_1_get_Current())
                    continue

                else:
                    return acc

                break

        if e.System_Collections_IEnumerator_MoveNext():
            return loop(e.System_Collections_Generic_IEnumerator_1_get_Current())

        else:
            raise Exception(SR_inputSequenceEmpty)


def reduce_back(folder: Callable[[_T, _T], _T], xs: IEnumerable_1[Any]) -> Any:
    arr: Array[_T] = to_array(xs)
    if len(arr) > 0:
        return reduce_back_1(folder, arr)

    else:
        raise Exception(SR_inputSequenceEmpty)


def replicate(n: int, x: Any) -> IEnumerable_1[Any]:
    def _arrow180(_arg: int, n: Any = n, x: Any = x) -> __A:
        return x

    return initialize(n, _arrow180)


def reverse(xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow181(__unit: None = None, xs: Any = xs) -> IEnumerable_1[_T]:
        return of_array(reverse_1(to_array(xs)))

    return delay(_arrow181)


def scan(folder: Callable[[_STATE, _T], _STATE], state: Any, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow182(__unit: None = None, folder: Any = folder, state: Any = state, xs: Any = xs) -> IEnumerable_1[_STATE]:
        acc: _STATE = state

        def mapping(x: _T | None = None) -> _STATE:
            nonlocal acc
            acc = folder(acc, x)
            return acc

        return concat(to_enumerable([singleton(state), map(mapping, xs)]))

    return delay(_arrow182)


def scan_back(folder: Callable[[_T, _STATE], _STATE], xs: IEnumerable_1[Any], state: Any) -> IEnumerable_1[Any]:
    def _arrow183(__unit: None = None, folder: Any = folder, xs: Any = xs, state: Any = state) -> IEnumerable_1[_STATE]:
        return of_array(scan_back_1(folder, to_array(xs), state, None))

    return delay(_arrow183)


def skip(count: int, source: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow184(__unit: None = None, count: Any = count, source: Any = source) -> IEnumerator[_T]:
        e: IEnumerator[_T] = of_seq(source)
        try:
            for _ in range(1, count + 1, 1):
                if not e.System_Collections_IEnumerator_MoveNext():
                    raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "source")

            def compensation(__unit: None = None) -> None:
                pass

            return Enumerator_enumerateThenFinally(compensation, e)

        except Exception as match_value:
            dispose_2(e)
            raise match_value

    return mk_seq(_arrow184)


def skip_while(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow185(__unit: None = None, predicate: Any = predicate, xs: Any = xs) -> IEnumerable_1[_T]:
        skipped: bool = True

        def f(x: _T | None = None) -> bool:
            nonlocal skipped
            if skipped:
                skipped = predicate(x)

            return not skipped

        return filter(f, xs)

    return delay(_arrow185)


def tail(xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    return skip(1, xs)


def take(count: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow186(__unit: None = None, count: Any = count, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow187(i: int, e: IEnumerator[_T], count: Any = count, xs: Any = xs) -> _T | None:
        if i < count:
            if e.System_Collections_IEnumerator_MoveNext():
                return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

            else:
                raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "source")

        else:
            return None

    def _arrow188(e_1: IEnumerator[_T], count: Any = count, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow186, _arrow187, _arrow188)


def take_while(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow189(__unit: None = None, predicate: Any = predicate, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow190(e: IEnumerator[_T], predicate: Any = predicate, xs: Any = xs) -> _T | None:
        return (
            some(e.System_Collections_Generic_IEnumerator_1_get_Current())
            if (
                predicate(e.System_Collections_Generic_IEnumerator_1_get_Current())
                if e.System_Collections_IEnumerator_MoveNext()
                else False
            )
            else None
        )

    def _arrow191(e_1: IEnumerator[_T], predicate: Any = predicate, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate(_arrow189, _arrow190, _arrow191)


def truncate(count: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow192(__unit: None = None, count: Any = count, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow193(i: int, e: IEnumerator[_T], count: Any = count, xs: Any = xs) -> _T | None:
        return (
            some(e.System_Collections_Generic_IEnumerator_1_get_Current())
            if (e.System_Collections_IEnumerator_MoveNext() if (i < count) else False)
            else None
        )

    def _arrow194(e_1: IEnumerator[_T], count: Any = count, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow192, _arrow193, _arrow194)


def zip(xs: IEnumerable_1[Any], ys: IEnumerable_1[Any]) -> IEnumerable_1[tuple[_T1, _T2]]:
    def _arrow195(x: _T1, y: _T2, xs: Any = xs, ys: Any = ys) -> tuple[_T1, _T2]:
        return (x, y)

    return map2(_arrow195, xs, ys)


def zip3(xs: IEnumerable_1[Any], ys: IEnumerable_1[Any], zs: IEnumerable_1[Any]) -> IEnumerable_1[tuple[_T1, _T2, _T3]]:
    def _arrow196(x: _T1, y: _T2, z: _T3, xs: Any = xs, ys: Any = ys, zs: Any = zs) -> tuple[_T1, _T2, _T3]:
        return (x, y, z)

    return map3(_arrow196, xs, ys, zs)


def collect(mapping: Callable[[_T], _COLLECTION], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow197(__unit: None = None, mapping: Any = mapping, xs: Any = xs) -> IEnumerable_1[_U]:
        return concat(map(mapping, xs))

    return delay(_arrow197)


def where(predicate: Callable[[_T], bool], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    return filter(predicate, xs)


def pairwise(xs: IEnumerable_1[Any]) -> IEnumerable_1[tuple[_T, _T]]:
    def _arrow198(__unit: None = None, xs: Any = xs) -> IEnumerable_1[tuple[_T, _T]]:
        return of_array(pairwise_1(to_array(xs)))

    return delay(_arrow198)


def split_into(chunks: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Array[Any]]:
    def _arrow199(__unit: None = None, chunks: Any = chunks, xs: Any = xs) -> IEnumerable_1[Array[_T]]:
        return of_array(split_into_1(chunks, to_array(xs)))

    return delay(_arrow199)


def windowed(window_size: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Array[Any]]:
    def _arrow200(__unit: None = None, window_size: Any = window_size, xs: Any = xs) -> IEnumerable_1[Array[_T]]:
        return of_array(windowed_1(window_size, to_array(xs)))

    return delay(_arrow200)


def transpose(xss: IEnumerable_1[Any]) -> IEnumerable_1[IEnumerable_1[Any]]:
    def _arrow201(__unit: None = None, xss: Any = xss) -> IEnumerable_1[IEnumerable_1[_T]]:
        return of_array(map_1(of_array, transpose_1(map_1(to_array, to_array(xss), None), None), None))

    return delay(_arrow201)


def sort_with(comparer: Callable[[_T, _T], int], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow202(__unit: None = None, comparer: Any = comparer, xs: Any = xs) -> IEnumerable_1[_T]:
        arr: Array[_T] = to_array(xs)
        sort_in_place_with(comparer, arr)
        return of_array(arr)

    return delay(_arrow202)


def sort(xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> IEnumerable_1[Any]:
    def _arrow203(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(x, y)

    return sort_with(_arrow203, xs)


def sort_by(projection: Callable[[_T], _U], xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> IEnumerable_1[Any]:
    def _arrow204(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(projection(x), projection(y))

    return sort_with(_arrow204, xs)


def sort_descending(xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> IEnumerable_1[Any]:
    def _arrow205(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(x, y) * -1

    return sort_with(_arrow205, xs)


def sort_by_descending(
    projection: Callable[[_T], _U], xs: IEnumerable_1[Any], comparer: IComparer_1[Any]
) -> IEnumerable_1[Any]:
    def _arrow206(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(projection(x), projection(y)) * -1

    return sort_with(_arrow206, xs)


def sum(xs: IEnumerable_1[Any], adder: IGenericAdder_1[Any]) -> Any:
    def _arrow207(acc: _T, x: _T, xs: Any = xs, adder: Any = adder) -> _T:
        return adder.Add(acc, x)

    return fold(_arrow207, adder.GetZero(), xs)


def sum_by(f: Callable[[_T], _U], xs: IEnumerable_1[Any], adder: IGenericAdder_1[Any]) -> Any:
    def _arrow208(acc: _U, x: _T, f: Any = f, xs: Any = xs, adder: Any = adder) -> _U:
        return adder.Add(acc, f(x))

    return fold(_arrow208, adder.GetZero(), xs)


def max_by(projection: Callable[[_T], _U], xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow209(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> _T:
        return y if (comparer.Compare(projection(y), projection(x)) > 0) else x

    return reduce(_arrow209, xs)


def max(xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow210(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> _T:
        return y if (comparer.Compare(y, x) > 0) else x

    return reduce(_arrow210, xs)


def min_by(projection: Callable[[_T], _U], xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow211(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> _T:
        return x if (comparer.Compare(projection(y), projection(x)) > 0) else y

    return reduce(_arrow211, xs)


def min(xs: IEnumerable_1[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow212(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> _T:
        return x if (comparer.Compare(y, x) > 0) else y

    return reduce(_arrow212, xs)


def average(xs: IEnumerable_1[Any], averager: IGenericAverager_1[Any]) -> Any:
    count: int = 0

    def folder(acc: _T, x: _T, xs: Any = xs, averager: Any = averager) -> _T:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, x)

    total: _T = fold(folder, averager.GetZero(), xs)
    if count == 0:
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")

    else:
        return averager.DivideByInt(total, count)


def average_by(f: Callable[[_T], _U], xs: IEnumerable_1[Any], averager: IGenericAverager_1[Any]) -> Any:
    count: int = 0

    def _arrow213(acc: _U, x: _T, f: Any = f, xs: Any = xs, averager: Any = averager) -> _U:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, f(x))

    total: _U = fold(_arrow213, averager.GetZero(), xs)
    if count == 0:
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")

    else:
        return averager.DivideByInt(total, count)


def permute(f: Callable[[int], int], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    def _arrow214(__unit: None = None, f: Any = f, xs: Any = xs) -> IEnumerable_1[_T]:
        return of_array(permute_1(f, to_array(xs)))

    return delay(_arrow214)


def chunk_by_size(chunk_size: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Array[Any]]:
    def _arrow215(__unit: None = None, chunk_size: Any = chunk_size, xs: Any = xs) -> IEnumerable_1[Array[_T]]:
        return of_array(chunk_by_size_1(chunk_size, to_array(xs)))

    return delay(_arrow215)


def insert_at(index: int, y: Any, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    is_done: bool = False
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    def _arrow216(__unit: None = None, index: Any = index, y: Any = y, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow217(i: int, e: IEnumerator[_T], index: Any = index, y: Any = y, xs: Any = xs) -> _T | None:
        nonlocal is_done
        if e.System_Collections_IEnumerator_MoveNext() if (True if is_done else (i < index)) else False:
            return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

        elif i == index:
            is_done = True
            return some(y)

        else:
            if not is_done:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

            return None

    def _arrow218(e_1: IEnumerator[_T], index: Any = index, y: Any = y, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow216, _arrow217, _arrow218)


def insert_many_at(index: int, ys: IEnumerable_1[Any], xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    status: int = -1
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    def _arrow219(
        __unit: None = None, index: Any = index, ys: Any = ys, xs: Any = xs
    ) -> tuple[IEnumerator[_T], IEnumerator[_T]]:
        return (of_seq(xs), of_seq(ys))

    def _arrow220(
        i: int, tupled_arg: tuple[IEnumerator[_T], IEnumerator[_T]], index: Any = index, ys: Any = ys, xs: Any = xs
    ) -> _T | None:
        nonlocal status
        e1: IEnumerator[_T] = tupled_arg[0]
        e2: IEnumerator[_T] = tupled_arg[1]
        if i == index:
            status = 0

        inserted: _T | None
        if status == 0:
            if e2.System_Collections_IEnumerator_MoveNext():
                inserted = some(e2.System_Collections_Generic_IEnumerator_1_get_Current())

            else:
                status = 1
                inserted = None

        else:
            inserted = None

        if inserted is None:
            if e1.System_Collections_IEnumerator_MoveNext():
                return some(e1.System_Collections_Generic_IEnumerator_1_get_Current())

            else:
                if status < 1:
                    raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

                return None

        else:
            return some(value_1(inserted))

    def _arrow221(
        tupled_arg_1: tuple[IEnumerator[_T], IEnumerator[_T]], index: Any = index, ys: Any = ys, xs: Any = xs
    ) -> None:
        dispose_2(tupled_arg_1[0])
        dispose_2(tupled_arg_1[1])

    return generate_indexed(_arrow219, _arrow220, _arrow221)


def remove_at(index: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    is_done: bool = False
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    def _arrow222(__unit: None = None, index: Any = index, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow223(i: int, e: IEnumerator[_T], index: Any = index, xs: Any = xs) -> _T | None:
        nonlocal is_done
        if e.System_Collections_IEnumerator_MoveNext() if (True if is_done else (i < index)) else False:
            return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

        elif e.System_Collections_IEnumerator_MoveNext() if (i == index) else False:
            is_done = True
            return (
                some(e.System_Collections_Generic_IEnumerator_1_get_Current())
                if e.System_Collections_IEnumerator_MoveNext()
                else None
            )

        else:
            if not is_done:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

            return None

    def _arrow224(e_1: IEnumerator[_T], index: Any = index, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow222, _arrow223, _arrow224)


def remove_many_at(index: int, count: int, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    def _arrow225(__unit: None = None, index: Any = index, count: Any = count, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow226(i: int, e: IEnumerator[_T], index: Any = index, count: Any = count, xs: Any = xs) -> _T | None:
        if i < index:
            if e.System_Collections_IEnumerator_MoveNext():
                return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

            else:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

        else:
            if i == index:
                for _ in range(1, count + 1, 1):
                    if not e.System_Collections_IEnumerator_MoveNext():
                        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "count")

            return (
                some(e.System_Collections_Generic_IEnumerator_1_get_Current())
                if e.System_Collections_IEnumerator_MoveNext()
                else None
            )

    def _arrow227(e_1: IEnumerator[_T], index: Any = index, count: Any = count, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow225, _arrow226, _arrow227)


def update_at(index: int, y: Any, xs: IEnumerable_1[Any]) -> IEnumerable_1[Any]:
    is_done: bool = False
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    def _arrow228(__unit: None = None, index: Any = index, y: Any = y, xs: Any = xs) -> IEnumerator[_T]:
        return of_seq(xs)

    def _arrow229(i: int, e: IEnumerator[_T], index: Any = index, y: Any = y, xs: Any = xs) -> _T | None:
        nonlocal is_done
        if e.System_Collections_IEnumerator_MoveNext() if (True if is_done else (i < index)) else False:
            return some(e.System_Collections_Generic_IEnumerator_1_get_Current())

        elif e.System_Collections_IEnumerator_MoveNext() if (i == index) else False:
            is_done = True
            return some(y)

        else:
            if not is_done:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

            return None

    def _arrow230(e_1: IEnumerator[_T], index: Any = index, y: Any = y, xs: Any = xs) -> None:
        dispose_2(e_1)

    return generate_indexed(_arrow228, _arrow229, _arrow230)


__all__ = [
    "SR_enumerationAlreadyFinished",
    "SR_enumerationNotStarted",
    "SR_inputSequenceEmpty",
    "SR_inputSequenceTooLong",
    "SR_keyNotFoundAlt",
    "SR_notEnoughElements",
    "SR_resetNotSupported",
    "Enumerator_noReset",
    "Enumerator_notStarted",
    "Enumerator_alreadyFinished",
    "Enumerator_Seq_reflection",
    "Enumerator_FromFunctions_1_reflection",
    "Enumerator_cast",
    "Enumerator_concat",
    "Enumerator_enumerateThenFinally",
    "Enumerator_generateWhileSome",
    "Enumerator_unfold",
    "index_not_found",
    "check_non_null",
    "mk_seq",
    "of_seq",
    "delay",
    "concat",
    "unfold",
    "empty",
    "singleton",
    "of_array",
    "to_array",
    "of_list",
    "to_list",
    "generate",
    "generate_indexed",
    "append",
    "cast",
    "choose",
    "compare_with",
    "contains",
    "enumerate_from_functions",
    "enumerate_then_finally",
    "enumerate_using",
    "enumerate_while",
    "filter",
    "exists",
    "exists2",
    "exactly_one",
    "try_exactly_one",
    "try_find",
    "find",
    "try_find_back",
    "find_back",
    "try_find_index",
    "find_index",
    "try_find_index_back",
    "find_index_back",
    "fold",
    "fold_back",
    "fold2",
    "fold_back2",
    "for_all",
    "for_all2",
    "try_head",
    "head",
    "initialize",
    "initialize_infinite",
    "is_empty",
    "try_item",
    "item",
    "iterate",
    "iterate2",
    "iterate_indexed",
    "iterate_indexed2",
    "try_last",
    "last",
    "length",
    "map",
    "map_indexed",
    "indexed",
    "map2",
    "map_indexed2",
    "map3",
    "read_only",
    "CachedSeq_1_reflection",
    "CachedSeq_1__Clear",
    "cache",
    "all_pairs",
    "map_fold",
    "map_fold_back",
    "try_pick",
    "pick",
    "reduce",
    "reduce_back",
    "replicate",
    "reverse",
    "scan",
    "scan_back",
    "skip",
    "skip_while",
    "tail",
    "take",
    "take_while",
    "truncate",
    "zip",
    "zip3",
    "collect",
    "where",
    "pairwise",
    "split_into",
    "windowed",
    "transpose",
    "sort_with",
    "sort",
    "sort_by",
    "sort_descending",
    "sort_by_descending",
    "sum",
    "sum_by",
    "max_by",
    "max",
    "min_by",
    "min",
    "average",
    "average_by",
    "permute",
    "chunk_by_size",
    "insert_at",
    "insert_many_at",
    "remove_at",
    "remove_many_at",
    "update_at",
]



from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1755() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Process.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Process", string_type), ("ArcProcess", string_type), ("name", string_type), ("executes_protocol", string_type), ("performer", string_type), ("date", string_type), ("previous_process", string_type), ("next_process", string_type), ("input", string_type), ("output", string_type), ("comments", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Process: str
    ArcProcess: str
    name: str
    executes_protocol: str
    performer: str
    date: str
    previous_process: str
    next_process: str
    input: str
    output: str
    comments: str

IContext_reflection = _expr1755

def _arrow1768(__unit: None=None) -> IEncodable:
    class ObjectExpr1756(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1757(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("https://bioschemas.org/")

    class ObjectExpr1758(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("bio:LabProcess")

    class ObjectExpr1759(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:name")

    class ObjectExpr1760(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("bio:executesLabProtocol")

    class ObjectExpr1761(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("bio:parameterValue")

    class ObjectExpr1762(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:agent")

    class ObjectExpr1763(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:endTime")

    class ObjectExpr1764(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:object")

    class ObjectExpr1765(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            return helpers_9.encode_string("sdo:result")

    class ObjectExpr1766(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
            return helpers_10.encode_string("sdo:disambiguatingDescription")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1756()), ("bio", ObjectExpr1757()), ("Process", ObjectExpr1758()), ("name", ObjectExpr1759()), ("executesProtocol", ObjectExpr1760()), ("parameterValues", ObjectExpr1761()), ("performer", ObjectExpr1762()), ("date", ObjectExpr1763()), ("inputs", ObjectExpr1764()), ("outputs", ObjectExpr1765()), ("comments", ObjectExpr1766())])
    class ObjectExpr1767(IEncodable):
        def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_11))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_11.encode_object(arg)

    return ObjectExpr1767()


context_jsonvalue: IEncodable = _arrow1768()

__all__ = ["IContext_reflection", "context_jsonvalue"]




import os
from typing import Awaitable


def read_all_text(file_name: str) -> str:
    with open(file_name, "r") as f:
        return f.read()


def read_all_text_async(file_name: str) -> Awaitable[str]:
    async def read_all_text_async():
        with open(file_name, "r") as f:
            return f.read()

    return read_all_text_async()


def read_all_lines(file_name: str) -> list[str]:
    with open(file_name, "r") as f:
        return [line.strip("\n") for line in f.readlines()]


def read_all_lines_async(file_name: str) -> Awaitable[list[str]]:
    async def read_all_lines_async():
        with open(file_name, "r") as f:
            return [line.strip("\n") for line in f.readlines()]

    return read_all_lines_async()


def read_all_bytes(file_name: str) -> bytes:
    with open(file_name, "rb") as f:
        return f.read()


def read_all_bytes_async(file_name: str) -> Awaitable[bytes]:
    async def read_all_bytes_async():
        with open(file_name, "rb") as f:
            return f.read()

    return read_all_bytes_async()


def write_all_text(file_name: str, text: str) -> None:
    with open(file_name, "w") as f:
        f.write(text)


def write_all_text_async(file_name: str, text: str) -> Awaitable[None]:
    async def write_all_text_async():
        with open(file_name, "w") as f:
            f.write(text)

    return write_all_text_async()


def write_all_lines(file_name: str, lines: list[str]) -> None:
    with open(file_name, "w") as f:
        f.write("\n".join(lines))


def write_all_lines_async(file_name: str, lines: list[str]) -> Awaitable[None]:
    async def write_all_lines_async():
        with open(file_name, "w") as f:
            f.write("\n".join(lines))

    return write_all_lines_async()


def write_all_bytes(file_name: str, bytes: bytes) -> None:
    with open(file_name, "wb") as f:
        f.write(bytes)


def write_all_bytes_async(file_name: str, bytes: bytes) -> Awaitable[None]:
    async def write_all_bytes_async():
        with open(file_name, "wb") as f:
            f.write(bytes)

    return write_all_bytes_async()


def delete(path: str) -> None:
    os.remove(path)


def move(source: str, destination: str) -> None:
    os.rename(source, destination)


def copy(source: str, destination: str) -> None:
    with open(source, "rb") as f:
        data = f.read()
    with open(destination, "wb") as f:
        f.write(data)


def exists(path: str) -> bool:
    return os.path.exists(path)


def replace(source_file_name: str, destination_file_name: str, destination_backup_file_name: str) -> None:
    with open(destination_file_name, "rb") as f:
        data = f.read()
    with open(destination_backup_file_name, "wb") as f:
        f.write(data)
    os.replace(source_file_name, destination_file_name)



from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.person import Person
from ...arctrl_core.Table.arc_table import ArcTable
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.template import (Organisation, Template)
from ...fable_library.array_ import map as map_1
from ...fable_library.date import to_string as to_string_1
from ...fable_library.seq import map
from ...fable_library.types import (to_string, Array)
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (and_then, succeed, string, object, IRequiredGetter, guid, resize_array, IOptionalGetter, IGetters, datetime_local, array as array_2)
from ...thoth_json_core.encode import seq
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..decode import Decode_datetime
from ..encode import date_time
from ..ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder)
from ..person import (encoder as encoder_1, decoder as decoder_2)
from .arc_table import (encoder, decoder as decoder_1, encoder_compressed, decoder_compressed)

__A_ = TypeVar("__A_")

def _arrow2808(arg: Organisation) -> IEncodable:
    value: str = to_string(arg)
    class ObjectExpr2807(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string(value)

    return ObjectExpr2807()


Template_Organisation_encoder: Callable[[Organisation], IEncodable] = _arrow2808

def cb(text_value: str) -> Decoder_1[Organisation]:
    return succeed(Organisation.of_string(text_value))


Template_Organisation_decoder: Decoder_1[Organisation] = and_then(cb, string)

def Template_encoder(template: Template) -> IEncodable:
    def _arrow2814(__unit: None=None, template: Any=template) -> IEncodable:
        value_1: str
        copy_of_struct: str = template.Id
        value_1 = str(copy_of_struct)
        class ObjectExpr2813(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value_1)

        return ObjectExpr2813()

    def _arrow2816(__unit: None=None, template: Any=template) -> IEncodable:
        value_3: str = template.Name
        class ObjectExpr2815(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_3)

        return ObjectExpr2815()

    def _arrow2820(__unit: None=None, template: Any=template) -> IEncodable:
        value_4: str = template.Description
        class ObjectExpr2819(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2819()

    def _arrow2824(__unit: None=None, template: Any=template) -> IEncodable:
        value_5: str = template.Version
        class ObjectExpr2823(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr2823()

    def mapping(person: Person, template: Any=template) -> IEncodable:
        return encoder_1(person)

    def mapping_1(oa: OntologyAnnotation, template: Any=template) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def mapping_2(oa_1: OntologyAnnotation, template: Any=template) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    values_3: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("id", _arrow2814()), ("table", encoder(template.Table)), ("name", _arrow2816()), ("description", _arrow2820()), ("organisation", Template_Organisation_encoder(template.Organisation)), ("version", _arrow2824()), ("authors", seq(map(mapping, template.Authors))), ("endpoint_repositories", seq(map(mapping_1, template.EndpointRepositories))), ("tags", seq(map(mapping_2, template.Tags))), ("last_updated", date_time(template.LastUpdated))])
    class ObjectExpr2833(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], template: Any=template) -> Any:
            def mapping_3(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_3, values_3)
            return helpers_4.encode_object(arg)

    return ObjectExpr2833()


def _arrow2851(get: IGetters) -> Template:
    def _arrow2838(__unit: None=None) -> str:
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("id", guid)

    def _arrow2839(__unit: None=None) -> ArcTable:
        object_arg_1: IRequiredGetter = get.Required
        return object_arg_1.Field("table", decoder_1)

    def _arrow2840(__unit: None=None) -> str:
        object_arg_2: IRequiredGetter = get.Required
        return object_arg_2.Field("name", string)

    def _arrow2841(__unit: None=None) -> str:
        object_arg_3: IRequiredGetter = get.Required
        return object_arg_3.Field("description", string)

    def _arrow2842(__unit: None=None) -> Organisation:
        object_arg_4: IRequiredGetter = get.Required
        return object_arg_4.Field("organisation", Template_Organisation_decoder)

    def _arrow2843(__unit: None=None) -> str:
        object_arg_5: IRequiredGetter = get.Required
        return object_arg_5.Field("version", string)

    def _arrow2844(__unit: None=None) -> Array[Person] | None:
        arg_13: Decoder_1[Array[Person]] = resize_array(decoder_2)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("authors", arg_13)

    def _arrow2845(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_15: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("endpoint_repositories", arg_15)

    def _arrow2847(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_17: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("tags", arg_17)

    def _arrow2850(__unit: None=None) -> Any:
        object_arg_9: IRequiredGetter = get.Required
        return object_arg_9.Field("last_updated", Decode_datetime)

    return Template.create(_arrow2838(), _arrow2839(), _arrow2840(), _arrow2841(), _arrow2842(), _arrow2843(), _arrow2844(), _arrow2845(), _arrow2847(), _arrow2850())


Template_decoder: Decoder_1[Template] = object(_arrow2851)

def Template_encoderCompressed(string_table: Any, oa_table: Any, cell_table: Any, template: Template) -> IEncodable:
    def _arrow2857(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        value_1: str
        copy_of_struct: str = template.Id
        value_1 = str(copy_of_struct)
        class ObjectExpr2856(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value_1)

        return ObjectExpr2856()

    def _arrow2859(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        value_3: str = template.Name
        class ObjectExpr2858(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_3)

        return ObjectExpr2858()

    def _arrow2861(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        value_4: str = template.Description
        class ObjectExpr2860(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2860()

    def _arrow2863(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        value_5: str = template.Version
        class ObjectExpr2862(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr2862()

    def mapping(person: Person, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        return encoder_1(person)

    def mapping_1(oa: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def mapping_2(oa_1: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow2865(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> IEncodable:
        value_1_1: str = to_string_1(template.LastUpdated, "O", {})
        class ObjectExpr2864(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_1_1)

        return ObjectExpr2864()

    values_3: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("id", _arrow2857()), ("table", encoder_compressed(string_table, oa_table, cell_table, template.Table)), ("name", _arrow2859()), ("description", _arrow2861()), ("organisation", Template_Organisation_encoder(template.Organisation)), ("version", _arrow2863()), ("authors", seq(map(mapping, template.Authors))), ("endpoint_repositories", seq(map(mapping_1, template.EndpointRepositories))), ("tags", seq(map(mapping_2, template.Tags))), ("last_updated", _arrow2865())])
    class ObjectExpr2866(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, template: Any=template) -> Any:
            def mapping_3(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_3, values_3)
            return helpers_5.encode_object(arg)

    return ObjectExpr2866()


def Template_decoderCompressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[Template]:
    def _arrow2878(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> Template:
        def _arrow2867(__unit: None=None) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("id", guid)

        def _arrow2868(__unit: None=None) -> ArcTable:
            arg_3: Decoder_1[ArcTable] = decoder_compressed(string_table, oa_table, cell_table)
            object_arg_1: IRequiredGetter = get.Required
            return object_arg_1.Field("table", arg_3)

        def _arrow2869(__unit: None=None) -> str:
            object_arg_2: IRequiredGetter = get.Required
            return object_arg_2.Field("name", string)

        def _arrow2870(__unit: None=None) -> str:
            object_arg_3: IRequiredGetter = get.Required
            return object_arg_3.Field("description", string)

        def _arrow2871(__unit: None=None) -> Organisation:
            object_arg_4: IRequiredGetter = get.Required
            return object_arg_4.Field("organisation", Template_Organisation_decoder)

        def _arrow2872(__unit: None=None) -> str:
            object_arg_5: IRequiredGetter = get.Required
            return object_arg_5.Field("version", string)

        def _arrow2873(__unit: None=None) -> Array[Person] | None:
            arg_13: Decoder_1[Array[Person]] = resize_array(decoder_2)
            object_arg_6: IOptionalGetter = get.Optional
            return object_arg_6.Field("authors", arg_13)

        def _arrow2874(__unit: None=None) -> Array[OntologyAnnotation] | None:
            arg_15: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
            object_arg_7: IOptionalGetter = get.Optional
            return object_arg_7.Field("endpoint_repositories", arg_15)

        def _arrow2876(__unit: None=None) -> Array[OntologyAnnotation] | None:
            arg_17: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
            object_arg_8: IOptionalGetter = get.Optional
            return object_arg_8.Field("tags", arg_17)

        def _arrow2877(__unit: None=None) -> Any:
            object_arg_9: IRequiredGetter = get.Required
            return object_arg_9.Field("last_updated", datetime_local)

        return Template.create(_arrow2867(), _arrow2868(), _arrow2869(), _arrow2870(), _arrow2871(), _arrow2872(), _arrow2873(), _arrow2874(), _arrow2876(), _arrow2877())

    return object(_arrow2878)


def Templates_encoder(templates: Array[Template]) -> IEncodable:
    def mapping(template: Template, templates: Any=templates) -> IEncodable:
        return Template_encoder(template)

    values: Array[IEncodable] = map_1(mapping, templates, None)
    class ObjectExpr2881(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], templates: Any=templates) -> Any:
            def mapping_1(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_1(mapping_1, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr2881()


Templates_decoder: Decoder_1[Array[Template]] = array_2(Template_decoder)

__all__ = ["Template_Organisation_encoder", "Template_Organisation_decoder", "Template_encoder", "Template_decoder", "Template_encoderCompressed", "Template_decoderCompressed", "Templates_encoder", "Templates_decoder"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...arctrl_rocrate.ldobject import LDRef
from ...fable_library.list import (singleton, FSharpList)
from ...fable_library.seq import map
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IRequiredGetter, string, IGetters)
from ...thoth_json_core.types import (Decoder_1, IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _arrow1882(decoders: IGetters) -> LDRef:
    def _arrow1881(__unit: None=None) -> str:
        object_arg: IRequiredGetter = decoders.Required
        return object_arg.Field("@id", string)

    return LDRef(_arrow1881())


decoder: Decoder_1[LDRef] = object(_arrow1882)

def encoder(r: LDRef) -> IEncodable:
    def _arrow1884(__unit: None=None, r: Any=r) -> IEncodable:
        value: str = r.Id
        class ObjectExpr1883(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr1883()

    values: FSharpList[tuple[str, IEncodable]] = singleton(("@id", _arrow1884()))
    class ObjectExpr1885(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], r: Any=r) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr1885()


__all__ = ["decoder", "encoder"]




import uuid

from .types import FSharpRef


def parse(string: str) -> uuid.UUID:
    return uuid.UUID(string.strip("()"))


def try_parse(string: str, def_value: FSharpRef[uuid.UUID]) -> bool:
    try:
        def_value.contents = parse(string)
        return True
    except Exception:
        return False


def to_string(guid: uuid.UUID) -> str:
    return str(guid)


def new_guid() -> uuid.UUID:
    return uuid.uuid4()


def guid_to_array(guid: uuid.UUID) -> bytearray:
    return bytearray(guid.bytes_le)


def array_to_guid(guid: bytearray) -> uuid.UUID:
    return uuid.UUID(bytes_le=bytes(guid))


__all__ = [
    "parse",
    "try_parse",
    "to_string",
    "new_guid",
    "guid_to_array",
    "array_to_guid",
]



from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.Helper.collections_ import ResizeArray_iter
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (length, singleton, initialize, map, FSharpList, of_array, empty, iterate_indexed, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.seq2 import List_distinct
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .sparse_table import (SparseTable_GetEmptyComments_3ECCA699, SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

def from_sparse_table(label: str, label_tsr: str, label_tan: str, matrix: SparseTable) -> FSharpList[OntologyAnnotation]:
    if (length(matrix.CommentKeys) != 0) if (matrix.ColumnCount == 0) else False:
        comments: Array[Comment] = SparseTable_GetEmptyComments_3ECCA699(matrix)
        def _arrow3447(__unit: None=None, label: Any=label, label_tsr: Any=label_tsr, label_tan: Any=label_tan, matrix: Any=matrix) -> OntologyAnnotation:
            return_val: OntologyAnnotation = OntologyAnnotation.create()
            return_val.Comments = comments
            return return_val

        return singleton(_arrow3447())

    else: 
        def _arrow3448(i: int, label: Any=label, label_tsr: Any=label_tsr, label_tan: Any=label_tan, matrix: Any=matrix) -> OntologyAnnotation:
            def mapping(k: str) -> Comment:
                return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, i)))

            comments_1: Array[Comment] = list(map(mapping, matrix.CommentKeys))
            return OntologyAnnotation(SparseTable__TryGetValue_11FD62A8(matrix, (label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (label_tsr, i)), SparseTable__TryGetValue_11FD62A8(matrix, (label_tan, i)), comments_1)

        return initialize(matrix.ColumnCount, _arrow3448)



def to_sparse_table(label: str, label_tsr: str, label_tan: str, designs: FSharpList[OntologyAnnotation]) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, of_array([label, label_tan, label_tsr]), None, length(designs) + 1)
    comment_keys: FSharpList[str] = empty()
    def action(i: int, d: OntologyAnnotation, label: Any=label, label_tsr: Any=label_tsr, label_tan: Any=label_tan, designs: Any=designs) -> None:
        i_1: int = (i + 1) or 0
        oa: dict[str, Any] = OntologyAnnotation.to_string_object(d, True)
        add_to_dict(matrix.Matrix, (label, i_1), oa["TermName"])
        add_to_dict(matrix.Matrix, (label_tan, i_1), oa["TermAccessionNumber"])
        add_to_dict(matrix.Matrix, (label_tsr, i_1), oa["TermSourceREF"])
        def f(comment: Comment, i: Any=i, d: Any=d) -> None:
            nonlocal comment_keys
            pattern_input: tuple[str, str] = Comment_toString(comment)
            n: str = pattern_input[0]
            comment_keys = cons(n, comment_keys)
            add_to_dict(matrix.Matrix, (n, i_1), pattern_input[1])

        ResizeArray_iter(f, d.Comments)

    iterate_indexed(action, designs)
    class ObjectExpr3450:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3449(x: str, y: str) -> bool:
                return x == y

            return _arrow3449

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3450())), matrix.ColumnCount)


def from_rows(prefix: str | None, label: str, label_tsr: str, label_tan: str, line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[OntologyAnnotation]]:
    labels: FSharpList[str] = of_array([label, label_tan, label_tsr])
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number) if (prefix is None) else SparseTable_FromRows_Z5579EC29(rows, labels, line_number, prefix)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(label, label_tsr, label_tan, tupled_arg[3]))


def to_rows(prefix: str | None, label: str, label_tsr: str, label_tan: str, designs: FSharpList[OntologyAnnotation]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    m: SparseTable = to_sparse_table(label, label_tsr, label_tan, designs)
    if prefix is None:
        return SparseTable_ToRows_759CAFC1(m)

    else: 
        return SparseTable_ToRows_759CAFC1(m, prefix)



__all__ = ["from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]

